# file generated by setuptools_scm
# don't change, don't track in version control
__version__ = version = '0.2.17'  # type: str
__version_tuple__ = version_tuple = (0, 2, 17)  # type: tuple[int | str, ...]
